/* Z80 Calc program */

/* Float format is A.BCDEF... x10^Exp */

/* Precision constants */
#define DIGITS 8
/* FLOATSIZE is DIGITS+2 */
#define FLOATSIZE 10
/* LONGBUFSIZE is DIGITS*2 */
#define LONGBUFSIZE 16

/* This should be halfway between 0 and -1 */
#define BIAS 0x8000
#define BYTESPERINT 2

#define NEGFLAG FLOATSIZE-2
#define EXPONENT FLOATSIZE-1

#define float(x) int x[FLOATSIZE]

/* Temporary floats used in calculation. */
float(fbufa);
int longbuf[LONGBUFSIZE];

main() {
  float(fa);
  float(fb);
  float(fr);

  puts("FreeCalc GPL 98 AZZ\n");

  putn(3);
  putc('\n');
  printint(123);
  putc('\n');

  clearbuf(fa,FLOATSIZE);
  fa[0] = 2;
  fa[EXPONENT] = BIAS;

  clearbuf(fb,FLOATSIZE);
  fb[0] = 3;
  /*  fb[1] = 1;
      fb[2] = 4;
      fb[3] = 1;
      fb[4] = 5;
      fb[5] = 9;
      fb[NEGFLAG] = 1; */
  fb[EXPONENT] = BIAS;

  puts("fa = ");
  printfloat(fa);
  puts("\nfb = ");
  printfloat(fb);

  addfloat(fa, fb, fr);

  puts("\nfr = ");
  printfloat(fr);

  puts("\n... done\n");

  printint(2+3);
  printint(fa[0]+fb[0]);

}

clearbuf(fl,size)
int *fl, size;
{
  int x;

  for (x=0; x<size; x++) fl[x]=0;
}

printfloat(fl)
int *fl;
{
  int x, exp, epos;

  epos = exp = abs(fl[EXPONENT]);

  if (fl[NEGFLAG]) putc('-');

  if (fl[EXPONENT]<BIAS) {
    epos = 0; /* to print 3.14159e-20 */
  }
  for (x=0; x<DIGITS; x++) {
    if (x==epos+1) putc('.');
    putn(fl[x]);
  }

  if ((exp >= DIGITS)||(fl[EXPONENT]<BIAS)) {
    putc('e');
    if (fl[EXPONENT]<BIAS) putc('-');
    printint(exp);
  }

  putc('\n');

}

printint(n)
int n;
{
  int flag;

  flag = 0;

#define DODIGIT(x) if ((n>=x)||flag) { putn((n-(n%x))/x); n=n%x; flag=1; }

  DODIGIT(10000);
  DODIGIT(1000);
  DODIGIT(100);
  DODIGIT(10);
  putn(n);
}

putn(n)
int n;
{
  putc('0' + n);
}

abs(n)
int n;
{
  if (n<BIAS)
    return BIAS-n;
  else
    return n-BIAS;
}

/* Add float fb on to fa, leaving result in fr. */

addfloat(fa,fb,fr)
int *fa, *fb, *fr;
{
  int *fl, *fh, x, y, ediff;

  fl = fa; fh = fb;
  if (fa[EXPONENT] > fb[EXPONENT]) {
    fh = fa;
    fl = fb;
  }

  printfloat(fh);
  printfloat(fl);

  ediff = fh[EXPONENT] - fl[EXPONENT];

  clearbuf(fr, FLOATSIZE);
  clearbuf(longbuf, LONGBUFSIZE);
  copybuf(longbuf+(BYTESPERINT*DIGITS), fh, DIGITS);

  /* dump digits as they are added? why is 2+3=8? */

  for(x=0; x<DIGITS; x++) {
    if ((x+ediff)<DIGITS) {
      y = DIGITS+x+ediff;
      printint(y);putc(',');printint(x);putc(']');
      printint(longbuf[y]);putc('+');printint(fl[x]);
      longbuf[y] = longbuf[y] + fl[x];
      putc('=');printint(longbuf[y]);putc('\n');
      if (longbuf[y]>9) {
	longbuf[y] = longbuf[y]-10;
	longbuf[y-1] = longbuf[y-1] + 1;
      }
    }
  }

  fr[EXPONENT] = fh[EXPONENT];

  /* This could be generalised: have a "alter value" variable which
     controls how many places to offset the copy and also how much
     to alter the exponent. */

  /* This is wrong. It should shift everything 1 to the right and
     alter the exponent. */
  if (longbuf[DIGITS-1]) puts("overflow error\n");

  copybuf(fr, longbuf+(BYTESPERINT*DIGITS), DIGITS);

}

/* subfloat uses the above, but reverses the sign of one of the
   arguments.
   */

subfloat(fa,fb,fr)
int *fa, *fb, *fr;
{
}

/* Copy fb over fa. */

copybuf(fa, fb, size)
int *fa, *fb, size;
{
  int x;
  
  for (x=0; x<size; x++) fa[x] = fb[x];
}
