
; Adam's floating calculator. Preprocess with M4 then compile with asl.

; fptostring is still screwed. (spurious zero)
	
	CPU Z80

; --- Preprocessor macros.







; --- Defines

cr	EQU 10
eos	EQU 0

PRECISION 	EQU   	6	;This is the floating PRECISION in digits.
FP_SIZE EQU     (PRECISION/2)+2 ;Size in bytes of a fp number.
DIGIT   EQU     PRECISION/2	;Number of packed BCD digits required.
FP_STACK      	EQU	20      ;Floating point stack size (x FP_SIZE).


; --- The code begins

	ORG 0h

	DI
	LD SP,1ffeh
	LD IX,string_welcome
	CALL puts

	CALL InitializeFP

	LD DE,string_pi
	LD HL,myfloat
	CALL StringToFP
	LD HL,myfloat
	CALL PushArgument

	LD DE,string_two
	LD HL,myfloat
	CALL StringToFP
	LD HL,myfloat
	CALL PushArgument

	CALL ComputeMult

	LD HL,myfloat
	CALL PopArgument
	
	LD DE,string_out
	LD HL,myfloat
	CALL FPToString

	LD IX,string_out
	CALL puts

	LD IX,string_goodbye
	CALL puts

hang:	
	; HALT
	JR hang

err:	LD IX,string_error
	CALL puts
	JR hang
	
dumpbuf:
	LD C,16
dbloop:
	LD A,(IX+0)
	INC IX
	DEC C
	JR NZ,dbloop
	RET

puts:	LD A,(IX+0)
	CP 0h
	RET Z
	OUT (0h),A
	INC IX
	JR puts

string_welcome: 	DB "Hello world",cr,eos
string_goodbye:		DB cr,"Goodbye world",cr,eos
string_error:		DB "Error",cr,eos
string_pi: 		DB "3.3",eos
string_two:		DB "1.1",eos
string_out:		DS PRECISION+8
   DS FP_SIZE-1
myfloat: DS 1h

; These emulate Gameboy CPU commands ldd and ldi, which were used in
; the original source. I didn't call the new commands ldd and ldi
; because ldd and ldi exist with different meanings in the Z80
; instruction set.

ldpdec	MACRO x,y
	LD x,y
	DEC HL
	ENDM

ldpinc	MACRO x,y
	LD x,y
	INC HL
	ENDM	

;**********************************************************
;*     Floating-Point Math Package for GameBoy or Z80     *
;**********************************************************

;last edit: 16-Apr-97
; by Jeff Frohwein

;Version 1.2 - First release


                              ;It should be an even number because the
                              ;floating point routines can't handle odd.
                              ;Increasing it's size increases PRECISION
                              ;but is slower & requires more ram usage.
                              ;IF YOU MODIFY THIS VALUE, MODIFY CONSTANTS
                              ;TABLE VALUES AS WELL.

;
;	floating point ram
;

;Used in Add, Subtract, Multiply, & Divide routines
; these were all DIGIT+1
hold1: DS DIGIT+1
hold2: DS DIGIT+1
hold3: DS DIGIT+1
hold4: DS DIGIT+1
hold5: DS DIGIT+1
hold6: DS DIGIT+1
hold7: DS DIGIT+1
hold8: DS DIGIT+1
erri: DS 1h                   ;error flag

buf: DS FP_SIZE         ;working buffer
sign    EQU     buf+DIGIT   ;sign bit
exp     EQU     buf+DIGIT+1 ;exponent

rctrl: DS 1h                  ;rounding control flag 1=msd
rdigi: DS 1h                  ;rounding digit
signd   EQU     hold1+DIGIT
expd    EQU     hold1+DIGIT+1

   DS FP_SIZE-1
fpsink: DS 1h
   DS FP_SIZE-1
ftemp: DS 1h
   DS FP_SIZE-1
ftem1: DS 1h
   DS FP_SIZE-1
ftem2: DS 1h

;Floating Point Stack Variables
astka: DS 2h			; Stack pointer
ASTKSZ EQU   FP_SIZE * FP_STACK
astkl: DS ASTKSZ		; Stack itself

;Used by StringToFP routine
adds: DS 2h
bcadd: DS 2h
opst: DS 1h	; Whether this is the left or right half of a nibble.
ecnt: DS 1h
bcs: DS DIGIT+2

;Used by FPToString routine
addt: DS 2h
fsign: DS 1h
abufsiz EQU PRECISION+5
abuf: DS abufsiz
expo: DS 1h
fes: DS 1h
infes: DS 1h
outptr: DS 2h

;Used by StringToFP & FPToString routines
cnsbuf: DS PRECISION+8
xsign: DS 1h

;Used by PopFix & CNS routine
miscW1: DS 2h

; Initialize values needed by math package.
InitializeFP:
        ld      a,2*PRECISION
        ld      (infes),a

        ld      hl,astkl+ASTKSZ+FP_SIZE-1
        ld      a,l
        ld      (astka),a
        ld      a,h
        ld      (astka+1),a
	ret
;
; Setup arguments for Add, Sub, Mult, & Div.
;
GetArgAddr:
        ld      a,(astka)
        ld      e,a
        ld      a,(astka+1)
        ld      d,a

        ld      hl,FP_SIZE        ;Pop one argument off stack
        add     hl,de

        ld      a,l
        ld      (astka),a
        ld      a,h
        ld      (astka+1),a
        ld      b,h
        ld      c,l             ;bc = ptr to arg1

        ex      de,hl           ;hl = ptr to arg2
        ld      d,b             ;de = ptr to arg1
        ld      e,c
        ret
;
; Return result of addition of top two stack arguments.
; Replace top two arguments on stack with result.
;
ComputeAdd:
        call    GetArgAddr
aadd1:	call	fadd		; These use hl and de as args.
        jr      fpetst
;
; Return result of subtraction of top two stack arguments.
; Replace top two arguments on stack with result.
; (In X = Y - Z, argument first put on stack is Y.)
;
ComputeSub:
        call    GetArgAddr
asub1:	call	fsub
        jr      fpetst
;
; Return result of multiplication of top two stack arguments.
; Replace top two arguments on stack with result.
; (In X = Y - Z, argument first put on stack is Y.)
;
ComputeMult:
        call    GetArgAddr
amul1:	call	fmul
        jr      fpetst
;
; Return result of division of top two stack arguments.
; Replace top two arguments on stack with result.
; (In X = Y / Z, argument first put on stack is Y.)
;
ComputeDiv:
        call    GetArgAddr
adiv1:	call	fdiv

fpetst: ld      a,(erri)     ;Error during calculation?
	or	a
        ret     z            ;No

        ld      a,(astka)   ;Zero result on underflow
        ld      l,a
        ld      a,(astka+1)
        ld      h,a
        ld      (hl),0
        ret
;
; Replace what's on stack with it's negative
;
ComputeNeg:
        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

        ld      a,(hl)
        or      a               ;is it 0?
        ret     z               ;yes, can't negate zero

        dec     hl              ;Invert sign of number
        ld      a,(hl)
	xor	1
        ld      (hl),a
        ret
;
; Replace what's on stack with it's absolute
;
ComputeAbs:
        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

        dec     hl
        ld      (hl),0
        ret
;
; Replace what's on stack with it's sign (1 or -1)
;
ComputeSign:
        ld      a,(astka)
        ld      e,a
        ld      a,(astka+1)
        ld      d,a
        dec     de
        ld      a,(de)
        inc     de
	or	a
	ld	hl,fpone
        jp      z,vcopy

	ld	hl,fpnone
	jp	vcopy
;
;	compute sin(x) x=top of argument stack
;	return result in place of x
;
ComputeSine:
        call    quadc        ;compute quadrant

        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	ld	d,h
	ld	e,l
	ld	bc,ftemp
	call	amul1	     ;ftemp=x*x
	pop	af
	push	af	     ;a=quadrant
	rra	
        jr      c,sin10      ;quad odd, compute cosine

;  compute x*p(x*x) -- sine
	ld	de,ftem1

        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	call	vcopy	     ;ftem1=x*x
	ld	bc,sinx
	call	poly	     ;p(x*x)
        call    PrepOp
	ld	hl,ftem1
	call	amul1	     ;x*p(x*x)

;   compute sign of result
; positive for quadrants 0,1. negative for 2,3
; negate above for negative arguments
sin5:	pop	af	     ;quadrant
	ld	b,a
	pop	af	     ;sign
	rlca		     ;sign, 2 to the 1st bit
	xor	b	     ;quadrant, maybe modified for negative arg.

        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	dec	hl	     ;ptr to sign
	sub	2
        cp      128
        ret     nc           ;quadrant 0 or 1
	inc	(hl)	     ;else set result negative
        or      a            ;clear carry flag
	ret

; compute p(x*x) -- cosine
sin10:	ld	bc,cosx
	call	poly	     ;p(x*x)
        jr      sin5
;
;    compute cos(x) x=top of argument stack
; return result in place of x
; cos(x)=sin(x+pi/2)
;
ComputeCosine:
        call    PrepOp
	ld	hl,pic2	     ;pi/2
	call	aadd1	     ;tos=tos+pi/2
        jp      ComputeSine
;
;    compute tan(x) x=top of argument stack
; return result in place of x
; tan(x)=sin(x)/cos(x)
;
ComputeTan:
        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

        call    PushArgument   ;push copy of x onto arg stack
        call    ComputeCosine  ;cos(x)
	ld	de,ftem2
        call    popa1          ;ftem2=cos(x)
        call    ComputeSine
        call    PrepOp
	ld	hl,ftem2
	jp	adiv1	     ;sin(x)/cos(x)
;
;    compute sqr(x) x=top of argument stack
; return result in place of x
;
ComputeSqrRoot:
        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	ld	de,ftemp
	call	vcopy	     ;save x in ftemp

; compute exponent of first guess as exponent of x/2
        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	ld	a,(hl)
	or	a
	ret	z	     ;x=0

        sub     128
        cp      128
        jr      nc,sqr5       ;negative exponent

	rrca
	and	127
        jr      sqr6

sqr5:	cpl	
	inc	a
	rrca	
	and	127
	cpl	
        inc     a
sqr6:
        add     a,128
        ldpdec     (hl),a

; test for negative argument
	ld	a,(hl)
	or	a
        jp      nz,Error     ;neg argument

; do newton iterations
; newguess =( x/oldguess + oldguess ) /2
	ld	a,6	     ;do 6 iterations
sqr20:	push	af	     ;set new iteration count
	ld	bc,ftem1
	ld	de,ftemp     ;ftemp is 'x'

        ld      a,(astka)    ;guess
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	call	adiv1	     ;ftem1=x/guess
	ld	de,ftem1

        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	ld	b,h
	ld	c,l
	call	aadd1	     ;tos=(x/guess)+guess
        call    PrepOp
	ld	hl,fptwo
	call	adiv1	     ;tos=(x/guess+guess)/2
	pop	af
	dec	a	     ;decrement count
        jr      nz,sqr20     ;do another iteration
	ret	

;Set carry flag & return

PopError:
        pop     af
Error:  scf
        ret
;
;   evaluate p(x) using horners method (x is in ftemp)
; coefficient list pointer is in bc
; result replaces number on top of argument stack (y)
poly:   ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

        ex      de,hl        ;de=ptr to y

	ld	h,b
	ld	l,c	     ;hl ptr to coefficient list
	call	vcopy	     ;y=first coefficient

; multiply by x
poly1:	push	hl	     ;save coeff list pointer
        call    PrepOp
	ld	hl,ftemp
	call	amul1	     ;y=y*x

; add next coeff
        call    PrepOp
	pop	hl
	push	hl	     ;hl=coeff. list pointer
	call	aadd1	     ;y=y+coeff.

; bump pointer to next coefficient
	pop	hl	     ;coeff. pointer
        ld      bc,0-FP_SIZE-1
	add	hl,bc	     ;next coef sign
        ldpinc     a,(hl)
        cp      128
        jr      c,poly1      ;process next coefficient
	ret		     ;negative sign (-1) - ends list
;
; prepare for operation
;
PrepOp: ld      a,(astka)
        ld      e,a
        ld      a,(astka+1)
        ld      d,a

	ld	b,d
	ld	c,e
	ret	
;
;     quadrant computation
; pops top of argument stack
; compute/gets sine of argument, quadrant of argument
; and index into quadrant
;
;    exits with:
; sp pointing to quadrant, mod 4
; sp+2 pointing to sign of argument
; top of argument stack has index into quadrant
quadc:  ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	dec	hl	     ;point to sign
	ld	b,(hl)
	xor	a
	ld	(hl),a	     ;arg. sign=0
	ld	h,b

        pop     de           ;pop return addr
        push    hl           ;put sign on stack
        push    de           ;push return

; compute quadrant of abs(x)
        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

        call    PushArgument ;put copy of arg. onto stack
        call    PrepOp
	ld	hl,pic1	     ;2/pi
	call	amul1	     ;tos=x*2/pi
        call    ComputeInt   ;tos=int(x*2/pi)

        ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

        call    PushArgument ;another copy
        call    PopFix       ;pop tos to de
	ld	a,e
	push	af	     ;quadrant
        call    PrepOp
	ld	hl,pic2
	call	amul1	     ;tos=int(x*2/pi)
	ld	de,ftemp
	call	popa1	     ;ftemp=tos
        call    PrepOp
	ld	hl,ftemp
	call	asub1	     ;tos=tos-ftemp
	pop	af	     ;a=quadrant, low order byte
	and	3	     ;mod 4
	pop	hl
	push	af	     ;save quadrant on stack
	jp	(hl)	     ;return

;
;   int function action routine
;
ComputeInt:
        call    PrepOp
aint:
	ld	a,(bc)
	sub	129
        cp      128
        jr      c,aint1

; zero if value less than one
	xor	a
	ld	(bc),a
	ret	

; exp > 0
aint1:  sub     PRECISION-1
	ret	nc
	ld	d,a	     ;count
	dec	bc
aint2:	dec	bc
	ld	a,(bc)
	and	0f0h
	ld	(bc),a
	inc	d
	ret	z
	xor	a
	ld	(bc),a
	inc	d
        jr      nz,aint2
	ret	
;
;   copys FP_SIZE bytes at addr hl to addr de
; on exit hl points to adr-1 of last byte copied
;
vcopy:  ld      c,FP_SIZE
vcop1:  LD A,(HL)
	ld	(de),a
	DEC HL
	dec	de
	dec	c
        jr      nz,vcop1
	ret	
;
; push value addr by hl onto arg stack
;
PushArgument:
        ld      e,l
        ld      d,h
psha1:  ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

        ld      bc,0-FP_SIZE
	add	hl,bc

        ld      a,l
        ld      (astka),a     ;dec arg stack pointer
        ld      a,h
        ld      (astka+1),a

        ex      de,hl         ;exchange de & hl

        jr      vcopy
;
;   pop arg stack
; hl contains addr to put popped value at
;
PopArgument:
        ex      de,hl
popa1:  ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a

	push	hl
        ld      bc,FP_SIZE
	add	hl,bc

        ld      a,l
        ld      (astka),a     ;inc stack pointer
        ld      a,h
        ld      (astka+1),a

	pop	hl
	jp	vcopy
;
; fix floating to positive integer
; return integer value in de
; fp value from top of arg stack, pop arg stack
;
PopFix: ld      a,(astka)
        ld      l,a
        ld      a,(astka+1)
        ld      h,a
	ld	b,h
	ld	c,l
	push	hl
	call	aint
	ld	hl,fpsink
        call    PopArgument
	pop	hl
	ld	c,(hl)	     ;exponent
	dec	hl
	ld	a,(hl)	     ;sign
	or	a
        jp      nz,Error     ;negative no good

        ld      de,0-FP_SIZE+1
	add	hl,de
	ld	de,0
	ld	a,c
	or	a
	ret	z

	dec	c	     ;set up for loop close test
pfix4:	inc	hl
	ld	a,(hl)
	rrca	
	rrca	
	rrca	
	rrca	
	call	mul10
        ret     c        ;error

	dec	c
        ld      a,c
        cp      128
        ccf
        ret     nc       ;return if C is positive

	ld	a,(hl)
	call	mul10
        ret     c        ;error

	dec	c
        ld      a,c
        cp      128
        jr      nc,pfix4 ;jump if C is negative

        or      a        ;clear carry flag
	ret
;
; take next digit in a (mask to 0fh), accumulate to de
;
mul10:  ld      a,l
        ld      (miscW1),a
        ld      a,h
        ld      (miscW1+1),a

	ld	h,d	     ;get original value in hl
	ld	l,e
	add	hl,hl	     ;double it
	ret	c
	add	hl,hl	     ;quaddruple it
	ret	c
	add	hl,de	     ;add original for result of 5 x
	ret	c
	add	hl,hl	     ;result is 10 x
	ret	c

        ex      de,hl

        ld      a,(miscW1)
        ld      l,a
        ld      a,(miscW1+1)
        ld      h,a

	and	0fh
        add     a,e
	ld	e,a
	ld	a,d
        adc     a,0            ;propogate the carry
	ld	d,a
	ret	

;Terminate ascii string
TermStr:
        ld      b,0
;Add ascii character in b to FP string output
chout:  push    af
        push    hl
        ld      a,(outptr)
        ld      l,a
        ld      a,(outptr+1)
        ld      h,a
        ld      (hl),b
        inc     hl
        ld      a,l
        ld      (outptr),a
        ld      a,h
        ld      (outptr+1),a
        pop     hl
        pop     af
        ret
;
;       Output fp number addr by hl
;       to an ascii string pointed to de.
;
FPToString:
        ld      a,e
        ld      (outptr),a
        ld      a,d
        ld      (outptr+1),a
        ld      bc,0-DIGIT-1
	add	hl,bc
	ld	b,h
	ld	c,l
	ld	hl,abuf	     ;output buffer
	ld	a,(infes)    ;output format
	ld	(fes),a	     ;store it
        ld      e,DIGIT
	ld	(hl),0	     ;clear round-off overflow buffer
	inc	hl	     ;abuf+1
;
nxt:	ld	a,(bc)	     ;get digit and unpack
	ld	d,a
	rra	
	rra	
	rra	
	rra	
	and	0fh	     ;remove bottom digit
        ldpinc     (hl),a       ;store top digit in output buffer (abuf)
	ld	a,d	     ;now get bottom digit
	and	0fh
        ldpinc     (hl),a       ;store it
	inc	bc
	dec	e
        jr      nz,nxt

	ld	a,(bc)
	ld	(fsign),a    ;store sign of number
	xor	a
	ld	(hl),a	     ;clear round-off buffer (abuf+13) 12 digit no rnd
	ld	hl,xsign     ;exponent sign store
	ld	(hl),a	     ;clear xsign
;
fix:	inc	bc	     ;get exponent
	ld	a,(bc)
	or	a	     ;exponent zero?
        jr      z,zro        ;yes

	sub	128	     ;remove normalizing bias
        jr      nz,fix2

	inc	(hl)	     ;inc xsign to negative flag (1)later zero

fix2:   cp      128
        jr      c,chk13

	cpl		     ;it's a negative exponent
	inc	(hl)	     ;inc xsign to negative (1)
zro:	inc	a
chk13:	ld	hl,expo	     ;exponent temp store
	ld	(hl),a
	ld	e,a
        cp      PRECISION
	ld	hl,fes	     ;format temp byte
        jr      c,chkxo

chk40:	ld	a,1	     ;force exponential printout
	or	(hl)	     ;set format for xout
	ld	(hl),a
;
chkxo:	ld	a,(hl)	     ;check if exponential printout
	rra	
        jr      nc,chkx3

	and	0fh
        cp      PRECISION
        jr      c,chkx2

        ld      a,PRECISION-1  ;max digits
chkx2:	ld	d,a
	inc	a
        jr      round
;
chkx3:	and	0fh	     ;add exponent & decimal places
	ld	d,a
        add     a,e
        cp      PRECISION+1
	ld	b,a
        jr      c,chkxn

	ld	a,(hl)
	and	40h
        jr      nz,chk40
;
chkxn:	ld	a,(xsign)    ;check exponent sign
	or	a
        jr      nz,xneg      ;it's negative

	ld	a,b
        jr      round
;
xneg:	ld	a,d	     ;sub exponent & decimal place count
	sub	e
        jr      nc,xn2

xn1:	ld	a,(infes)
        cp      128
        jp      c,zerons

	and	0eh
        jp      z,zerons

	rrca
	ld	e,a
	dec	e
	ld	c,1
	ld	hl,abuf-1
        jr      nrnd

xn2:    jr      z,xn1
        jr      round
;
;
clean:	ld	b,1fh	     ;clear flags
	and	b
        cp      PRECISION+1
	ret	c

        ld      a,PRECISION+1  ;max digits out
	ret	
;
; this routine is used to round data to the
; specified decimal place
round:	call	clean
	ld	c,a
	ld	b,0
	ld	hl,abuf+1
	add	hl,bc	     ;get round-off addr

        ld      a,l
        ld      (addt),a
        ld      a,h
        ld      (addt+1),a

	ld	a,(hl)
	cp	5	     ;round if >=5
        jr      c,trl1
;
less1:	dec	hl
	inc	(hl)	     ;round up
	ld	a,(hl)
	or	a
        jr      z,trl2

	cp	10	     ;check if rounded number >9
        jr      nz,trail

	ld	(hl),0
        jr      less1
;
; this routine eliminates trailing zeros
trail:  ld      a,(addt)
        ld      l,a
        ld      a,(addt+1)
        ld      h,a

trl1:   dec     hl
trl2:	ld	a,(fes)	     ;check if trailing zeros are wanted
	rla	
        jr      c,fprnt      ;yes, go print data

trl3:	ld	a,(hl)
	or	a	     ;is it a zero?
        jr      nz,fprnt     ;no, go print

	dec	hl
	dec	c	     ;yes, fix output digit count
        ld      a,c
        cp      128
        jp      nc,zeron     ;jump if C is negative

        jr      trl3
;
; print format routines
fprnt:	ld	hl,abuf
	ld	a,(hl)	     ;check if rounded up to 1
	or	a
        jr      z,nrnd       ;jump if not

	ld	b,1
	ld	a,(xsign)    ;is exponent negative?
	or	a
        jr      z,posr

	ld	b,0-1
;
posr:	ld	a,(expo)     ;get exponent
	or	a
        jr      nz,po2       ;is it zero? (e+0)

	ld	(xsign),a
	ld	b,1
po2:    add     a,b          ;fix exponent count
	ld	(expo),a
	inc	e
	inc	c
	dec	hl
;
nrnd:	inc	hl
	ld	a,c
        cp      PRECISION+1    ;check for maximum digits out
        jr      nz,nrnd1

	dec	c
nrnd1:	ld	a,(fsign)    ;check if neg #
	rra	
        jr      nc,prin2     ;go output radix & number

	call	neg	     ;output (-)
        jr      pri21
;
prin2:	call	space	     ;output a space
pri21:	ld	a,(fes)	     ;get output format
	rra		     ;check if exponential output format
        jr      c,xprin

	ld	a,(xsign)    ;get exp sign
	or	a	     ;check if neg exp
        jr      z,posit

	ld	a,c
	or	a
        jr      nz,prin4     ;output radix & number

        jp      zerons       ;no digits after radix, output zero & done
;
prin4:	call	radix	     ;print decimal point
prin6:  xor     a
	or	e
        jr      z,prin5      ;jump if no zeros to print

	call	zero	     ;force print a zero
	dec	e
        jr      nz,prin6
;
prin5:	call	nout	     ;print ascii digit
        jr      nz,prin5
        jp      TermStr 
;
posit:	call	nout
	dec	e	     ;bump exp count
        jr      nz,posit

	ld	a,c	     ;check if more digits to output
	or	a
        jp      z,TermStr    ;no, done
        cp      128
        jp      nc,TermStr

        jr      prin4        ;now print decimal point
;
; exponential format output
xprin:	call	nout
        jr      z,ndec       ;integer?

	call	radix	     ;no. print decimal point
xpri2:	call	nout
        jr      nz,xpri2
;
ndec:	ld	b,'e'	     ;print 'e'
	call	chout
	ld	a,(xsign)
	or	a
        jr      z,xpri3

	call	neg	     ;print exp sign (-)
	ld	a,(expo)
	inc	a
        jr      xout2

xpri3:	ld	b,'+'	     ;exp (+)
	call	chout
;
; convert the exponent from binary-to-ascii
; and print the result.
xout:	ld	a,(expo)
	dec	a
xout2:	ld	c,100		;  print the 100s place
	ld	d,0
	call	conv
	cp	'0'	     ;skip leading zeros
        jr      z,xo21

	inc	d
	call	chout
xo21:	ld	a,e
	ld	c,10		; now the 10s place
	call	conv
	cp	'0'
        jr      nz,xo3

	dec	d
        jr      nz,xo4

xo3:	call	chout
xo4:	ld	a,e
        add     a,'0'          ;add ascii bias
	ld	b,a
        jp      zeroc
;
conv:	ld	b,'0'-1
conv1:  inc     b
	sub	c
        jr      nc,conv1

        add     a,c
	ld	e,a
	ld	a,b
	ret	
;
; change bcd digit to ascii & print
nout:	ld	a,(hl)
        add     a,'0'
	ld	b,a
	call	chout
	inc	hl
	dec	c	     ;dec total digits printed count
	ret	

; print fp zero
zeron:  ld      b,' '
zeroc:  call    chout
zerons: call    zero
        jp      TermStr
;
; common symbol loading routines
neg:	ld	b,'-'
	jp	chout
zero:	ld	b,'0'
	jp	chout
space:	ld	b,' '
	jp	chout
radix:	ld	b,'.'
	jp	chout

; ---

StringToFP:
	ld	b,0
	ld	a,(de)
	cp	'+'	     ;look for leading plus or minus on input
        jr      z,stofp2

	cp	'-'
        jr      nz,stofp3

	ld	b,1		; b is presumably the mantissa sign
stofp2: inc     de		; Move past the leading + or -
stofp3: push    bc
	push	hl
        call    SToFP         ;input fp number
	pop	hl
	dec	hl
        pop     bc
        ld      (hl),b		; Write the mantissa sign in.
        ret

; converts fp string at de, update de past terminator
; puts terminator in b, puts fp # at addr in hl.
; Sets carry if valid number not found.

SToFP:  push    hl
        ld      l,e
        ld      h,d

        dec     hl

        ld      a,l
        ld      (adds),a
        ld      a,h
        ld      (adds+1),a
        call    ibscn           ;get first non-space
        cp      '&'
        jr      z,fpin6

        dec     hl
        call    ibscn2          ;add back to buffer
        call    fpins
        pop     hl
        jp      nc,entr3
        ret

; --- this all deals with getting *hex* number from input

fpin6:  call    ibscn           ;get 'h'
        cp      'h'             ;is it hex?
        jp      nz,PopError        ;no
	call    getnib
        jp      c,PopError         ;bad hex number
	ld      e,a
        ld      d,0

        ld      b,4
fpin7:  call    getnib
        jp      c,fpin8

        dec     b
        jp      z,PopError         ;overflow

        push    hl              ;de = de * 16
        ld      l,e
        ld      h,d
        add     hl,hl
        add     hl,hl
        add     hl,hl
        add     hl,hl
        ld      e,l
        ld      d,h
        pop     hl

        add     a,e             ;add a to de
        ld      e,a
        ld      a,0
        adc     a,d
        ld      d,a

        jr      fpin7

fpin8:  push    hl

        ld      l,e             ;put hex number in hl
        ld      h,d

        ld      de,cnsbuf       ;convert it to a ascii decimal string
        call    cns
        ld      a,CR
        ld      (de),a

        ld      de,cnsbuf-1
        ld      a,e
        ld      (adds),a
        ld      a,d
        ld      (adds+1),a
        call    fpins
        pop     de
        pop     hl
        push    de
        call    entr3
        pop     de
        ld      a,(de)
        ld      b,a
        inc     de
        or      a          ;clear carry flag
        ret

getnib: call    ibscn		; get the next byte
        sub     '0'
; These are required because asl doesn't like immediate expressions
; with chars in.
; This seems a slightly longwinded way of comparing with 10!
FUDGEA	EVAL '9'
FUDGEA  EVAL FUDGEA + 1
FUDGEA	EVAL FUDGEA - '0'
        cp      FUDGEA
        ccf
        ret     nc		; if greater then 10, return
FUDGEB EVAL 'a'
FUDGEB EVAL FUDGEB - '0'
        sub     FUDGEB
FUDGEC EVAL 'f'
FUDGEC EVAL FUDGEC + '1'
FUDGEC EVAL FUDGEC - 'a'
FUDGEC EVAL FUDGEC - '0'
        cp      FUDGEC
        ccf
        ret     c
        add     a,10
        ret

; --- this is the important bit that does the FP string to FP.

fpins:  push    de

        xor     a
        ld      (opst),a
        ld      (ecnt),a
        ld      (fsign),a

        ld      hl,bcs       ;clear temporary storage areas & bc buffer
        ld      c,DIGIT+2
	call	clear

scanc:	ld	de,0
        ld      hl,bcs       ;bc=pack buffer
scan0:  ld      a,l          ;pack buffer pointer
        ld      (bcadd),a
        ld      a,h
        ld      (bcadd+1),a
scanp:	ld	hl,scanp
	push	hl	     ;used for return from other routines
	xor	a
	ld	(xsign),a    ;clear exp sign byte
;
scang:	call	ibscn
	jr      c,scanx      ;found a number, go pack it
	cp	'.'	     ;radix?
        jr      z,scan5      ;process radix pointers
	cp	'e'	     ;exp?
	jp	z,excon	     ;found 'e'', go process exp #

;this char not legal in #
	ld	b,a	     ;move terminator to b
	ld	a,(opst)     ;check if any digits yet
	and	10h
	jp	nz,entr2

;legal fp number not found
fpin1:	pop	hl	     ;rid of scanp link
        pop     de           ;text pointer
	scf	
	ret	

;found decimal point
scan5:	xor	a	     ;found radix process radix pointers for exp
	or	d	     ;any digits yet?
        jr      nz,scan6

        add     a,0c0h       ;set ecnt - stop counting digits
	or	e	     ;no int digits, bit 7 is count (or don't) flag
	ld	e,a	     ;bit 6 is negative exp flag
	ret	

scan6:	ld	a,80h	     ;set ecnt to count digits
	or	e
	ld	e,a
	ret	
;
scanx:	and	0fh	     ;found number - remove ascii bias
	ld	b,a
	ld	hl,opst	     ;set first char flag
	ld	a,30h
	or	(hl)
	ld	(hl),a
	xor	a
	or	b	     ;is char zero?
        jr      nz,pack

	or	d	     ;leading zero? ie; any int digits?
        jr      nz,pack

	or	e
	ld	e,a
	ret	z	     ;if counting yet,
	inc	e	     ;ecnt+1-count zeros for exp count
	ret	
;
; bcd pack digits into pair bc
;
pack:	ld	a,e
	rla	
        jr      c,pack1

	inc	e
pack1:	ld	a,e
	ld	(ecnt),a     ;digit count for exp count
	inc	d	     ;total digit count (d has top/bot flag bit 7)
	ld	a,d
	and	7fh	     ;remove top/bot flag
        cp      PRECISION+1  ;limit input digits
	ret	nc

        ld      a,d
        cp      128
        jr      nc,botm
;
top:	or	80h	     ;set msb for top flag
        ld      d,a

        ld      a,(bcadd)    ;get bc addr
        ld      l,a
        ld      a,(bcadd+1)
        ld      h,a

        ld      a,b
	rlca
	rlca	
	rlca	
	rlca	
; This writes the number (in B). into bcs
	ld	(hl),a	     ;save char in bc
	ret	
;
botm:	and	7fh	     ;strip msb (bottom flag)
	ld	d,a
	ld	a,b

        ld      a,(bcadd)
        ld      l,a
        ld      a,(bcadd+1)
        ld      h,a

        ld      a,b
	or	(hl)	     ;or in top number
	ldpinc     (hl),a       ;put number back in bc
	pop	bc
	jp	scan0

; -- ibscn: get next char, set C if number.

ibscn:  ld      a,(adds)     ;input buffer pointer
        ld      l,a
        ld      a,(adds+1)
        ld      h,a

ibscn1: inc     hl           ;get next byte
	ld	a,(hl)
	cp	' '
        jr      z,ibscn1

	; azz: stupid bug???
ibscn2: PUSH AF
	ld      a,l
        ld      (adds),a
        ld      a,h
        ld      (adds+1),a
	POP AF

; check for ascii numbers
nmchk:	cp	'9'+1
	ret	nc
	cp	'0'
	ccf	
	ret	
;
; adjust a number in bc buffer & return value
entr2:	ld	de,0
ent1:	push	bc	     ;terminator
	call	fixe	     ;normalize floating point #
	pop	bc	     ;terminator
	pop	de	     ;scanp link
	pop	de	     ;old text addr
        or      a
        ret

; This bit copies the result from bcs back to where it should be.
entr3:  ld      e,l
        ld      d,h
        ld      c,DIGIT+2
        ld      hl,bcs+DIGIT+1
	call	vcopy

        ld      a,(adds)
        ld      l,a
        ld      a,(adds+1)
        ld      h,a

        ex      de,hl

	inc	de
	or	a
	ret

; clear storage areas
; hl = starting address
; c = count
clear:	xor	a
clear1: ldpinc     (hl),a
	dec	c
        jr      nz,clear1
	ret	
;
; convert ascii exponent of number in the input buffer
; to binary. normalize exponent according to the input
; format of the number.
excon:	call	ibscn	     ;get character
        jr      c,exc3

;        cp      plsrw        ;check for unary sign
;        jr      z,exc4

	cp	'+'
        jr      z,exc4

;        cp      minrw
;        jr      z,exc2

	cp	'-'
        jr      nz,fperr     ;no sign or number?

exc2:	ld	a,1
	ld	(xsign),a    ;save sign
exc4:	call	ibscn
        jr      nc,fperr     ;no number?

exc3:	call	ascdc	     ;convert ascii to binary
        jr      ent1         ;normalize # & return
;
; convert ascii to binary
; three consecutive numbers <128 may be converted
ascdc:  ex      de,hl

	ld	hl,0
asc1:	ld	a,(de)	     ;get chr from input buffer, no spaces allowed
	call	nmchk	     ;check if #
        jr      nc,asc2
	sub	'0'	     ;remove ascii bias
	ld	b,h
	ld	c,l
	add	hl,hl
	add	hl,hl
	add	hl,bc
	add	hl,hl
	ld	c,a
	ld	b,0
	add	hl,bc
	inc	de
        jr      asc1

asc2:   ex      de,hl

	ld	b,a	     ;save terminator

        ld      a,l
        ld      (adds),a     ;save ibuf addr
        ld      a,h
        ld      (adds+1),a

	ld	a,d
	or	a
        jr      nz,fperr     ;too big >255

	ld	a,e
	rla	
        jr      c,fperr      ;too big >127

	rra
	ret	

fperr:	pop	bc	     ;ascdc ret link
	jp	fpin1
;
; normalize input buffer
fixe:   ex      de,hl

        ld      a,(bcs)
	or	a	     ;is it zero?
        jr      z,zz2

	call	chkpn	     ;set exp pos/neg
        add     a,80h          ;add exp bias
zz2:    ld      (bcs+DIGIT+1),a;store normalized exp in bc
	ret	
;
chkpn:	ld	a,(ecnt)     ;get exp count-set in 'scan' routine
	ld	e,a
	and	3fh	     ;strip bits 7&8
	ld	b,a
	ld	a,(xsign)
	or	a
        jr      z,lpos       ;exponent is positive

	inc	h	     ;set sign in h
	ld	a,40h	     ;l is neg
	and	e	     ;check if e is negative
        jr      z,epos

	ld	a,l	     ;both e&l neg
	ld	l,b
        call    bpos1
	cpl	
	inc	a
	ret		     ;back to fixe
;
epos:	ld	a,l	     ;e&l neg
epos1:  cpl 
	inc	a
        add     a,b
	ret		     ;to fixe
;
lpos:	ld	a,40h	     ;exponent positive
	and	e	     ;is e negative?
        jr      z,bpos

	ld	a,b
	ld	b,l
        jr      epos1
;
bpos:	ld	a,b	     ;e&l pos
bpos1:  add     a,l
        cp      128
        ret     c

	pop	hl
        jr      fperr
;
; convert integer to string
; de = addr of string
; hl = value to convert
; exit:	de = updated value
;
cns:
	xor	a	     ;set for no leading zeroes
	ld	bc,0-10000
	call	rsub
	ld	bc,0-1000
	call	rsub
	ld	bc,0-100
	call	rsub
	ld	bc,0-10
	call	rsub
	ld	bc,0-1
	call	rsub
	ret	nz
	ld	a,'0'
	ld	(de),a
	inc	de
	ret	
;
; Take value in hl sub # in bc the
; most possible times.
; Put value on string at de.
; If a=0 then don't put zero on string.
; Return non-zero if a put on string
;
rsub:	push	de
	ld	d,0-1
rsub1:  ld      a,l
        ld      (miscW1),a
        ld      a,h
        ld      (miscW1+1),a
        inc     d
        add     hl,bc
        jr      c,rsub1

        ld      a,(miscW1)
        ld      l,a
        ld      a,(miscW1+1)
        ld      h,a
	ld	b,d
	pop	de
	or	b	     ;a gets 0 if a was 0 and b is 0
	ret	z

	ld	a,'0'
        add     a,b
	ld	(de),a
	inc	de
	ret	

;       Four Function Floating Point BCD
;
;		bc = de # hl
;	      # is +,0-,*, or /.
;	<bc>=address of result
;	<de>=address of 1st argument
;	<hl>=address of 2nd argument
; All addresses on entry point to the exponent part of #.
; Each # consists of PRECISION packed decimal digits,
; a sign, and a biased binary exponent. The exponent range
; is 10**-127 to 10**127. The number 0 is represented by
; the exponent 0. Positive numbers are indicated by the
; sign 0, negative numbers have a sign 1. The numbers are
; stored in memory as two BCD digits per byte starting at
; the low order address. All numbers are assumed to be normalized.

;
;	floating point addition
;
fadd:	push	bc
	call	expck	     ;fetch arguments
	ld	c,0
adsum:	dec	de

        ex      de,hl

        ld	a,(sign)
	xor	(hl)	     ;form sign of result
	ld	b,a

        ex      de,hl

	ld	a,(de)
	dec	de
	xor	c
	ld	(sign),a
	ld	hl,rctrl     ;rounding control flag
        ldpinc     a,(hl)
	or	a
	ld	a,(hl)	     ;get rounding digit
        jr      z,ads8

	rlca
	rlca	
	rlca	
	rlca	
ads8:   add     a,0b0h       ;force carry if digit > 5
	ld	a,b
	rra	
        jr      c,ads1       ;have sub

	rla		     ;restore carry
	call	add0	     ;perform addition
        jr      nc,ads2

	ld	b,4
	call	right
	ld	hl,exp
	inc	(hl)	     ;inc exp
	jp	z,over

ads2:	pop	bc	     ;get results addr
        jp      store        ;save results

zerex:	pop	hl
        jr      ads2

add0:   ld      hl,buf+DIGIT-1
        ld      b,DIGIT
add1:	ld	a,(de)
        adc     a,(hl)
	daa	
        ldpdec     (hl),a
	dec	de
	dec	b
        jr      nz,add1

	ret	nc
	inc	(hl)
	ret	
;
;	floating point subtraction
;
fsub:	push	bc
	call	expck	     ;get arguments
	ld	a,(sign)
	xor	1	     ;complement sign
	ld	(sign),a
        jr      adsum

ads1:	rla		     ;restore carry
	ccf		     ;complement for rounding
	call	sub0	     ;subtract arguments
	ld	hl,sign
        jr      c,ads4

	ld	a,(hl)	     ;get sign
	xor	1	     ;complement
	ld	(hl),a
ads7:	dec	hl
        ld      b,DIGIT
ads3:	ld	a,9ah
        sbc     a,(hl)       ;complement result
        add     a,0
	daa	
        ldpdec     (hl),a
	dec	b
	ccf	
        jr      nz,ads3

ads4:	ld	hl,buf
        ld      bc,DIGIT
ads5:	ld	a,(hl)
	or	a
        jr      nz,ads6

	inc	hl
	inc	b
	inc	b
	dec	c
        jr      nz,ads5

	xor	a
	ld	(exp),a
        jr      ads2

ads6:	cp	10h
        jr      nc,ads9

	inc	b
ads9:	ld	hl,exp
	ld	a,(hl)
	sub	b
	jp	z,under
	jp	c,under

	ld	(hl),a
	ld	a,b
	rlca	
	rlca	
	ld	b,a
	call	left
        jr      ads2

sub0:   ld      hl,buf+DIGIT-1
        ld      b,DIGIT
sub1:	ld	a,99h
        adc     a,0
	sub	(hl)

        ex      de,hl

        add     a,(hl)
	daa

        ex      de,hl

        ldpdec     (hl),a
	dec	de
	dec	b
        jr      nz,sub1
	ret	
;
;	floating point multiply
;		de and hl point to the args
fmul:	push	bc
	ld	a,(hl)
	or	a	     ;argument = 0?
        jr      z,fmul1a	; AZZ: these were fmul1+2

	ld	a,(de)
	or	a	     ;argument = 0?
        jr      z,fmul1a

        add     a,(hl)       ;form result exponent
        jr      c,fmovr
        cp      128
        jp      c,under         ;jump if A is positive

        jr      fmul1

fmovr:
        cp      128
        jp      nc,over

fmul1:	sub	128	     ;remove excess bias
fmul1a:	ld	(exp),a	     ;save exponent
	dec	de
	dec	hl
	ld	a,(de)

	xor	(hl)	     ;form result sign
	dec	hl
	dec	de
	push	hl
	ld	hl,sign	     ;get sign addr
        ldpdec     (hl),a       ;save sign
	xor	a
        ld      b,DIGIT+2
fmul2:  ldpdec     (hl),a       ;zero working buffer
	dec	b
        jr      nz,fmul2

	ld	a,(exp)
	or	a
	jp	z,zerex

        ld      c,DIGIT
        ld      hl,hold1+DIGIT
; get multiplier into holdpincng register
fmul3:	ld	a,(de)
        ldpdec     (hl),a       ;put in register
	dec	de
	dec	c
        jr      nz,fmul3

	ld	(hl),c
	dec	hl
	ld	b,250	     ;set loop count
fmul4:  ld      de,DIGIT+1
	ld	c,e
	add	hl,de

        ex      de,hl

	add	hl,de	     ;hl=next holdpincng register
	inc	b
        ld      a,b
        cp      128
        jr      c,fmul8      ;finished

fmul5:	ld	a,(de)	     ;get digits
        adc     a,a          ;times 2
	daa	
        ldpdec     (hl),a       ;put in holdpincng register
	dec	de
	dec	c
        jr      nz,fmul5

	inc	b	     ;inc loop count
        jr      nz,fmul4

; form 10x by adding 8x & 2x
; first get 8x

	inc	hl
	ld	de,hold5     ;next holdpincng register
        ld      c,DIGIT+1
	ld	b,c
fmul6:  ldpinc     a,(hl)
	ld	(de),a
	inc	de
	dec	c
        jr      nz,fmul6

        ld      hl,hold2+DIGIT  ;get 2x
	dec	de
fmul7:	ld	a,(de)
        adc     a,(hl)          ;form 10x
	daa	
	ld	(de),a
	dec	de
	dec	hl
	dec	b
        jr      nz,fmul7

	ld	b,249

        ex      de,hl

        jr      fmul4

fmul8:  ex      de,hl

	inc	hl

        ld      (hl),DIGIT+1 ;set next loop count
; perform accumulation of product
fmul9:	pop	bc	     ;get multiplier
        ld      hl,hold8+DIGIT+1
	dec	(hl)	     ;dec loop count
        jr      z,fmu14      ;finished

	ld	a,(bc)
	dec	bc
	push	bc
	dec	hl

        ex      de,hl

fmu10:  add     a,a          ;check for bit in carry
        jr      c,fmu11      ;found a bit
        jr      z,fmu12      ;zero, finished this digit

        ld      hl,0-DIGIT-1
	add	hl,de	     ;point to next holdpincng register

        ex      de,hl

        jr      fmu10

fmu11:	ld	c,a
	or	a	     ;clear carry
	call	add0	     ;accumulate product
	ld	a,(de)
        add     a,(hl)
	daa	
	ld	(hl),a
	ld	a,c
	dec	de
        jr      fmu10

; rotate right 1 byte
fmu12:	ld	b,8
	call	right
        jr      fmul9

fmu14:	ld	a,(buf)
	and	0f0h	     ;check if normalized
        jr      z,fmu17

	ld	a,d
	and	0f0h
	ld	hl,sign-1
        jr      fmu18

fmu17:	ld	b,4
	ld	hl,exp
	dec	(hl)
	jp	z,under

	call	left	     ;normalize
	ld	a,d	     ;get digit shifted off
; perform rounding
	rrca	
	rrca	
	rrca	
	rrca	
fmu18:	cp	50h
        jr      c,fmu16

	inc	a
	and	0fh
        ld      c,DIGIT
fmu15:  adc     a,(hl)
	daa	
        ldpdec     (hl),a
	ld	a,0
	dec	c
        jr      nz,fmu15

; check for rounding overflow
	jp	nc,ads2	     ;no overflow

	inc	hl
	ld	(hl),10h
	ld	hl,exp
	inc	(hl)
	jp	nz,ads2
	jp	over

; rounding not needed
fmu16:	and	0fh
        add     a,(hl)
	ld	(hl),a
	jp	ads2
;
;	floating point division
;
fdiv:	push	bc
	ld	a,(hl)	     ;fetch divisor exp
	or	a	     ;divide by 0?
	jp	z,divz

	ld	a,(de)
	or	a	     ;dividend = 0?
	jp	z,insp

	sub	(hl)
        jr      c,divun
        cp      128
        jp      nc,over
        jr      fdi1

divun:  cp      128
        jp      c,under         ;jump if positive

fdi1:   add     a,129        ;form quotient exp
	ld	(expd),a

        ex      de,hl

	push	de
	call	load	     ;fetch dividend
	pop	de

        ex      de,hl

	ld	a,(sign)
	dec	hl
	xor	(hl)	     ;form quotient sign
	ld	(signd),a

        ex      de,hl

	dec	de
	ld	bc,hold1
div0:   ld      l,PRECISION
div1:	push	bc
	push	hl
	ld	c,0	     ;quotient digit = 0
div3:	scf		     ;set carry
        ld      hl,buf+DIGIT-1
        ld      b,DIGIT
div4:	ld	a,99h
        adc     a,0

        ex      de,hl

	sub	(hl)

        ex      de,hl

        add     a,(hl)
	daa	
        ldpdec     (hl),a
	dec	de
	dec	b
        jr      nz,div4

	ld	a,(hl)
	ccf	
        sbc     a,0
	ld	(hl),a
	rra	
        ld      hl,DIGIT
	add	hl,de

        ex      de,hl

	inc	c	     ;inr quotient
	rla	
        jr      nc,div3

	or	a	     ;clear carry
	call	add0	     ;restore dividend
        ld      hl,DIGIT
	add	hl,de

        ex      de,hl

	push	bc
	ld	b,4
	call	left	     ;shift dividend
	pop	bc
	dec	c
	pop	hl
	ld	h,c
	pop	bc
	ld	a,l
        jr      nz,div5

        cp      PRECISION
        jr      nz,div5

	ld	hl,expd
	dec	(hl)
	call	z,under
        jr      div0

div5:	rra
	ld	a,h
        jr      nc,div6

	ld	a,(bc)
	rlca	
	rlca	
	rlca	
	rlca	
        add     a,h
	ld	(bc),a	     ;store quotient
	inc	bc
        jr      div7

div6:	ld	(bc),a	     ;store quotient
div7:	dec	l	     ;dec digit count
        jr      nz,div1

	ld	hl,expd
	pop	bc
        jr      storo

; fetch & align arguments for
; addition & subtraction
expck:	ld	a,(de)
	sub	(hl)	     ;difference of exps
	ld	c,0
        jr      nc,expc1

	inc	c

        ex      de,hl

	cpl	
	inc	a
expc1:	ld	b,a
	ld	a,(de)
	ld	(exp),a
	ld	a,b
        cp      PRECISION
        jr      c,expc2

        ld      a,PRECISION
expc2:	rlca	
	rlca	
	ld	b,a
	and	4
	ld	(rctrl),a    ;set rounding control
	push	bc
	push	de
	call	load	     ;load smaller value
        ld      a,8*DIGIT+16
	sub	b
        cp      8*DIGIT+16
        jr      z,expc3

	and	0f8h
	rra	
	rra	
	rra	
        add     a,e
	ld	e,a
	ld	a,d
        adc     a,0
	ld	d,a
	ld	a,(de)	     ;get rounding digit
	ld	(rdigi),a    ;save
expc3:	call	right	     ;align values
	pop	de
	pop	bc
	ret	

; load argument into buffer
load:   ld      de,sign
        ld      c,DIGIT+1
        dec     hl
load1:  ldpdec     a,(hl)
        ld      (de),a
        dec     de
        dec     c
        jr      nz,load1

        xor     a
        ld      (de),a
        dec     de
        ld      (de),a
        ld      (rdigi),a    ;zero rounding digit
        ret 

; store results in memory
store:  ld      hl,exp
storo:  ld      e,DIGIT+2
stor1:  ldpdec     a,(hl)
	ld	(bc),a
	dec	bc
	dec	e
        jr      nz,stor1
	ret	

; shift right number of digits in b/4
right:  ld      c,DIGIT+1
righ1:	ld	hl,buf-1
	ld	a,b
	sub	8	     ;check if byte can be shifted
        jr      nc,righ3

	dec	b
        push    af
        ld      a,b
        cp      128
        jr      c,righ5

        pop     af
        ret

righ5:
        pop     af

	or	a
righ2:	ld	a,(hl)
	rra	
        ldpinc     (hl),a
	dec	c
        jr      nz,righ2
        jr      right

; shift right one byte
righ3:	ld	b,a
	xor	a
righ4:	ld	d,(hl)
        ldpinc     (hl),a
	ld	a,d
	dec	c
        jr      nz,righ4
        jr      right

; shift left number of digits in b/4
left:   ld      c,DIGIT+1
	ld	hl,sign-1
lef1:	ld	a,b
	sub	8
        jr      nc,lef3

	dec	b
        push    af
        ld      a,b
        cp      128
        jr      c,lef5

        pop     af
        ret

lef5:
        pop     af

	or	a
lef2:	ld	a,(hl)
	rla	
        ldpdec     (hl),a
	dec	c
        jr      nz,lef2
        jr      left

; shift left one byte
lef3:	ld	b,a
	xor	a
lef4:	ld	d,(hl)
        ldpdec     (hl),a
	ld	a,d
	dec	c
        jr      nz,lef4
        jr      left

; set flags for overflow, underflow
; and divide by zero
over:
        jp      Error

under:  ld      a,0ffh
	ld	(erri),a
insp:	inc	sp
	inc	sp
	ret	

divz:
        jp      Error

;
;***** Floating Point Constants *****
;
; Note: If floating point PRECISION is changed
;       the following table MUST be modified as well.

;
;	sine coefficient list
;
        DB    -1           ;marks end of sine coefficient list
        DB    0

        DB    10h
        DB    00h
        DB    00h
        DB    0
fpone:  DB    128+1        ;.100000 e 1

        DB    16h
        DB    66h
        DB    67h
        DB    1
        DB    128          ;-.166667 e 0 (-1/3)

        DB    83h
        DB    33h
        DB    33h
        DB    0
        DB    128-2        ;.833333 e-2 (1/5)

        DB    19h
        DB    84h
        DB    13h
        DB    1
        DB    128-3        ;-.198413 e-3 (-1/7)

        DB    27h
        DB    55h
        DB    73h
        DB    0
        DB    128-5        ;.275573 e-5 (1/9)

        DB    25h
        DB    05h
        DB    21h
        DB    1
sinx:   DB    128-7        ;-.250521 e-7 (-1/11)

;
;	cosine coefficient list
;
        DB    -1           ;marks end of cosine coefficient list
        DB    0

        DB    10h
tmp:    DB    00h
        DB    00h
        DB    0
one:    DB    128+1        ;.100000 e 1 (1/1)

        DB    50h
        DB    00h
        DB    00h
        DB    1
matub:  DB    128          ;-.500000 e 0 (-1/2)

        DB    41h
        DB    66h
        DB    67h
        DB    0
rands:  DB    128-1        ;.416667 e-1 (1/4)

        DB    13h
        DB    88h
        DB    89h
        DB    1
        DB    128-2        ;.138889 e-2 (-1/6)

        DB    24h
        DB    80h
        DB    16h
        DB    0
        DB    128-4        ;.248016 e-4 (1/8)

        DB    27h
        DB    55h
        DB    73h
        DB    1
cosx:   DB    128-6        ;.275573 e-6 (-1/10)

        DB    10h
        DB    00h
        DB    00h
        DB    1
fpnone: DB    128+1        ;-1.0

        DB    20h
        DB    00h
        DB    00h
        DB    0
fptwo:  DB    128+1        ;2.0

        DB    15h
        DB    70h
        DB    80h
        DB    0
pic2:   DB    128+1        ;pi/2 .157080 e 1

        DB    63h
        DB    66h
        DB    20h
        DB    0
pic1:   DB    128          ;2/pi .636620 e 0

;lcstka: DW    cstkl

        DB    13h
        DB    10h
        DB    72h
        DB    0
snd2:   DB    128+6        ;131072.0

        END

